package sos.connection;

import java.util.GregorianCalendar;
import java.util.Properties;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import java.sql.Driver;
import java.sql.Statement;
import java.sql.Connection;

import sos.util.SOSClassUtil;
import sos.util.SOSLogger;
import sos.util.SOSString;

/**
 * <p>Title: </p>
 * <p>Description: Implementierung der SOSCOnnection-Klasse</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: SOS GmbH</p>
 * @author <a href="mailto:andreas.pueschel@sos-berlin.com">Andreas Pschel</a>
 * @author <a href="mailto:ghassan.beydoun@sos-berlin.com">Ghassan Beydoun</a>
 * @resource postgresql*jdbc3.jar sos.util.jar
 */

public class SOSPgSQLConnection extends sos.connection.SOSConnection {

    /** Umsetzung fr %lcase, %ucase, %now */
    private static final String replacement[] = { "LOWER", "UPPER", "CURRENT_TIMESTAMP",
            "FOR UPDATE"};

    /**
     * @param connection
     * @param logger
     * @throws java.lang.Exception
     */
    public SOSPgSQLConnection(Connection connection, SOSLogger logger)
            throws Exception {
        super(connection, logger);
        super.replacement = replacement;
        //prepare( connection) ;
    }

    /**
     * @param connection
     * @throws java.lang.Exception
     */
    public SOSPgSQLConnection(Connection connection) throws Exception {
        super(connection);
        super.replacement = replacement;
        prepare(connection);
    }

    /**
     * Konstruktor fr SOSOracleConnection
     * 
     * @param configFileName
     * @param logger
     *            logging Objekt
     * @throws java.lang.Exception
     * @see #SOSOracleConnection(String)
     * @see #SOSOracleConnection(String, String, String, String, SOSLogger)
     * @see #SOSOracleConnection(String, String, String, String)
     */
    public SOSPgSQLConnection(String configFileName, SOSLogger logger)
            throws Exception {

        super(configFileName, logger);
        super.replacement = replacement;
    }

    /**
     * Konstruktor fr SOSPgSQLConnection
     * 
     * @param configFileName
     * @throws java.lang.Exception
     * @see #SOSOracleConnection(String , SOSLogger)
     * @see #SOSOracleConnection(String, String, String, String, SOSLogger)
     * @see #SOSOracleConnection(String, String, String, String)
     */
    public SOSPgSQLConnection(String configFileName) throws Exception {
        super(configFileName);
        super.replacement = replacement;
    }

    /**
     * Konstruktor
     * 
     * @param driver
     * @param url
     * @param dbuser
     * @param dbpassword
     * @throws java.lang.Exception
     */
    public SOSPgSQLConnection(String driver, String url, String dbuser,
            String dbpassword, SOSLogger logger) throws Exception {
        super(driver, url, dbuser, dbpassword, logger);
        super.replacement = replacement;
    }

    /**
     * Konstruktor
     * 
     * @param driver
     * @param url
     * @param dbuser
     * @param dbpassword
     * @throws java.lang.Exception
     * @see #SOSPgSQLConnection(String , SOSLogger)
     * @see #SOSPgSQLConnection(String)
     * @see #SOSPgSQLConnection(String, String, String, String, SOSLogger)
     *  
     */
    public SOSPgSQLConnection(String driver, String url, String dbuser,
            String dbpassword) throws Exception {
        super(driver, url, dbuser, dbpassword);
        super.replacement = replacement;
    }

    /**
     * stellt die Datenbankverbindung her
     * 
     * @throws java.lang.Exception
     *             falls ein Datenbankfehler vorliegt.
     */
    public void connect() throws Exception {

        Properties properties = new Properties();
        logger.debug6("calling " + SOSClassUtil.getMethodName());

        if (SOSString.isEmpty(url))
                throw new Exception(SOSClassUtil.getMethodName()
                        + ": missing database url.");
        if (SOSString.isEmpty(driver))
                throw new Exception(SOSClassUtil.getMethodName()
                        + ": missing database driver.");
        if (SOSString.isEmpty(dbuser))
                throw new Exception(SOSClassUtil.getMethodName()
                        + ": missing database user.");
        //if (SOSString.isEmpty(dbpassword))
        //        throw new Exception(SOSClassUtil.getMethodName()
        //                + ": missing database password.");
        if (SOSString.isEmpty(dbpassword)) dbpassword = "";

        properties.setProperty("user", dbuser);
        properties.setProperty("password", dbpassword);

        Driver driver = (Driver) Class.forName(this.driver).newInstance();
        connection = driver.connect(url, properties);
        if (connection == null)
                throw new Exception("can't connect to database");
        logger.debug6(".. successfully connected to " + url);

        prepare(connection);

    }

    /**
     * prepare: Aufbereiten nach connect
     * 
     * @param connection
     *            falls vorhanden ist
     * @throws java.lang.Exception
     *             falls ein Datenbankfehler vorliegt.
     */
    private void prepare(Connection connection) throws Exception {

        logger.debug6("calling " + SOSClassUtil.getMethodName());
        Statement stmt = null;

        try {
            if (connection == null)
                    throw new Exception("can't connect to database");

            connection.setAutoCommit(false);
            connection.rollback();
            stmt = connection.createStatement();
            String NUMERIC_CHARACTERS = "SELECT set_config('lc_numeric', '', true)";
            String DATE_STYLE = "SELECT set_config('datestyle', 'ISO, YMD', true)";
            String DEFAULT_TRANSACTION_ISOLATION = "SELECT set_config('default_transaction_isolation', 'repeatable read', true)";
            /*
            stmt.addBatch(NUMERIC_CHARACTERS);
            stmt.addBatch(DATE_STYLE);
            stmt.addBatch(DEFAULT_TRANSACTION_ISOLATION);
            stmt.executeBatch();
            */
            stmt.execute(NUMERIC_CHARACTERS);
            logger.debug9(".. " + NUMERIC_CHARACTERS + " successfully set.");
            stmt.execute(DATE_STYLE);
            logger.debug9(".. " + DATE_STYLE + " successfully set.");
            stmt.execute(DEFAULT_TRANSACTION_ISOLATION);
            logger.debug9(".. " + DEFAULT_TRANSACTION_ISOLATION + " successfully set.");
        } catch (Exception e) {
            throw e;
        } finally {
            try {
                if (stmt != null) stmt.close();
            } catch (Exception e) {
            }
        }
    }


    /**
     * liefert PgSQL timestamp Funktion zur�ck.
     * 
     * @param dateString
     * @return oracle timestamp Funktion
     * @throws java.lang.Exception
     */
    public String toDate(String dateString) throws Exception {
        if (SOSString.isEmpty(dateString))
                throw new Exception(SOSClassUtil.getMethodName()
                        + ": dateString has no value.");
        return "TO_TIMESTAMP('" + dateString + "','YYYY-MM-DD HH24:MI:SS')";
    }


    /**
     * Aktuelles Datum als GregorianCalendar
     * 
     * @return
     */
    protected GregorianCalendar getDateTime(String format) throws Exception {
        GregorianCalendar gc = new GregorianCalendar();

        String timestamp = this.getSingleValue("select CURRENT_TIMESTAMP");

        if (timestamp.length() > 19) {
            timestamp = timestamp.substring(0, 19);
        }

        java.util.Date date = sos.util.SOSDate.getDate(timestamp, format);

        gc.setTime(date);

        return gc;
    }
    
    /**
     * 
     * @param inputString
     * @return
     */
    protected String pseudoFunctions( String inputString) throws Exception {
        
        logger.debug6("Calling " + SOSClassUtil.getMethodName());

        Pattern pattern = Pattern.compile(CAST_PATTERN);
        Matcher matcher = pattern.matcher(inputString);
        StringBuffer buffer = new StringBuffer();
        String replaceString = null;
        String token;
       
        logger.debug9("..inputString [" + inputString + "]");

        while ((matcher.find())) {

            replaceString = matcher.group().toLowerCase();
            if ( matcher.group(1) != null &&  matcher.group(6) != null) {
                token = matcher.group(6).replaceFirst("\\)","").trim();

                if ( token.matches(".*varchar.*")) {
                  replaceString = replaceString.replaceAll("varchar",",'999999999999999999')");
                  replaceString = replaceString.replaceFirst("%cast","TRIM(TO_CHAR");
                } else if ( token.matches(".*character.*")) {
                  replaceString = replaceString.replaceAll("character",",'999999999999999999'");
                  replaceString = replaceString.replaceFirst("%cast","TO_CHAR");
                } else if (token.matches(".*integer.*")) {
                  replaceString = replaceString.replaceAll("integer",",'999999999999999999'");
                  replaceString = replaceString.replaceFirst("%cast","TO_NUMBER");
                }
            } // if
            if ( matcher.group(3) != null && matcher.group(4) != null) { // group 4 "VALUE <data_type>"
                token = matcher.group(4).replaceFirst("\\(","").trim();
                
                if ( token.matches(".*varchar.*")) {
                   replaceString = replaceString.replaceAll("varchar",",'999999999999999999')");
                   replaceString = replaceString.replaceAll("%cast","TRIM(TO_CHAR");
                } else if ( token.matches(".*character.*")) {
                   replaceString = replaceString.replaceAll("character",",'999999999999999999'");
                   replaceString = replaceString.replaceAll("%cast","TO_CHAR");
                } else if (token.matches(".*integer.*")) {
                   replaceString = replaceString.replaceAll("integer",",'999999999999999999'");
                   replaceString = replaceString.replaceAll("%cast","TO_NUMBER");
               }
            }
            replaceString = replaceString.toUpperCase();
            matcher.appendReplacement(buffer, replaceString);
        } // while
        matcher.appendTail(buffer);

        logger.debug6(".. result [" + buffer.toString() + "]");
        return buffer.toString();
    } // pseudoFunctions    

}