// $Id: z_posix_process.cxx 4293 2006-05-24 18:12:03Z jz $

#include "zschimmer.h"
#ifdef Z_UNIX

#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/fcntl.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <signal.h>

#include "log.h"
#include "z_process.h"

using namespace std;


namespace zschimmer {

//-------------------------------------------------------------------------kill_process_immediately
// S.a. z_windows_process.cxx

void kill_process_immediately( int pid, bool kill_descendants )
{
    Z_LOG( "kill(" << pid << ",SIGKILL)\n" );
    int err = ::kill( kill_descendants? -pid : pid, SIGKILL );
    if( err )  throw_errno( errno, "kill" );
}

//-------------------------------------------------------------------------------------------------

namespace posix {

const int idle_priority         = +16;
const int below_normal_priority = +6;
const int normal_priority       = 0;
const int above_normal_priority = -6;
const int high_priority         = -16;

//-----------------------------------------------------------------------------priority_from_string
// Siehe auch Process::priority_class()

int priority_from_string( const string& priority )
{
    int result;

    if( priority.length() >= 1  &&  isdigit( (uint)priority[ 0 ] ) )
    {
        return as_int( priority );
    }
    else
    {
        string p = lcase( priority );
        
        if( p == "idle"         )  result = idle_priority;
        else
        if( p == "below_normal" )  result = below_normal_priority;
        else                                                                                                  
        if( p == "normal"       )  result = normal_priority;
        else
        if( p == "above_normal" )  result = above_normal_priority;
        else
        if( p == "high"         )  result = high_priority;
        else
        {
            throw_xc( "Z-4011", priority );
        }
    }

    return result;
}

//---------------------------------------------------------------------------argv_from_command_line
    
vector<string> argv_from_command_line( const string& command_line )
{
    vector<string>       argv;
    const unsigned char* p    = (const unsigned char*)command_line.c_str();

    while(1)
    {
        //fprintf( stderr, "p=%s\n", p );
        while( isspace( p[0] ) )  p++;
        if( p[0] == '\0' )  break;

        string arg;
        arg.reserve( 1000 );

        if( p[0] != '"'  &&  p[0] != '\'' )
        {
            while( p[0] != '\0'  &&  !isspace( p[0] )  &&  p[0] != '"'  &&  p[0] != '\'' )  arg += *p++;
        }

        if( p[0] == '"' )
        {
            p++;

            while( p[0] != '\0'  &&  p[0] != '"' )
            {
                if( p[0] == '\\' ) 
                {
                    p++;
                    if( p[0] == '\0' )  break;
                }

                arg += *p++;
            }

            if( p[0] == '"' )  p++;
        }
        else
        if( p[0] == '\'' )
        {
            p++;

            while( p[0] != '\0'  &&  p[0] != '\'' )
            {
                arg += *p++;
            }

            if( p[0] == '\'' )  p++;
        }

        //fprintf(stderr,"%d=%s.\n", argv.size(), arg.c_str() );
        argv.push_back( arg );
    }

    return argv;
}

//---------------------------------------------------------------------------------Process::Process
    
Process::Process()
:
    _zero_(this+1)
{
}

//--------------------------------------------------------------------------------Process::~Process
    
Process::~Process()
{
    try
    {
        close();
    }
    catch( exception& x ) { Z_LOG( "Process::close(): " << x << "\n" ); }
}

//-----------------------------------------------------------------------------------Process::close

void Process::close()
{
    if( _working_pid )  call_waitpid( false );          // Um Zombie zu beenden

    //_pid = 0;
    _working_pid = 0;
    _exit_code = 0;
    _termination_signal = 0;
    _terminated = false;
}

//-----------------------------------------------------------------------------------Process::start

void Process::start( const string& command_line )
{
#   ifndef SYSTEM_WINDOWS
        Z_LOG( "signal(SIGCHLD,SIG_DFL)\n" );
        ::signal( SIGCHLD, SIG_DFL );                 // Java verndert das Signal-Verhalten, so dass waitpid() ohne diesen Aufruf versagte.
#   endif


    start( argv_from_command_line( command_line ) );
}

//-----------------------------------------------------------------------------------Process::start

void Process::start( const vector<string>& command )
{
    if( _working_pid )  throw_xc( "Z-PROCESS-002" );
    if( command.size()      == 0 )  throw_xc( "Z-4009" );
    if( command[0].length() == 0 )  throw_xc( "Z-4009" );

    if( _command_line == "" )  _command_line = join( " ", command );

    
    //Z_LOG( "fork(), execv( \"" << command << " )\n" );
    Z_LOG( "fork()\n" );

    _working_pid = fork();
    if( _working_pid == 0 )
    {
        if( _priority_set )  setpriority( PRIO_PROCESS, getpid(), _priority );

        int n = sysconf( _SC_OPEN_MAX );
        for( int i = 3; i < n; i++ )  ::close(i);
        //::close( STDIN_FILENO );
        int new_stdin = ::open( "/dev/null", O_RDONLY );
        if( new_stdin != -1  &&  new_stdin != STDIN_FILENO )  dup2( new_stdin, STDIN_FILENO ),  ::close( new_stdin );



        Z_FOR_EACH( Environment, _environment, e )  set_environment_variable( e->first, e->second );


        S log;
        log << "execv(";

        const char** argv = new const char*[ command.size() + 1 ];
        for( int i = 0; i < command.size(); i++ )
        {
            argv[i] = command[i].c_str();
            log << ",\"" << argv[i] << "\"";
        }
        argv[ command.size() ] = NULL;

        log << ")\n";


        execvp( argv[0], (char**)argv );
        
        fprintf( stderr, "execv errno=%d %s\ncommand=%s", errno, strerror(errno), ((string)log).c_str() );
        _exit(99);
    }
    else
    if( _working_pid == -1 )  
        throw_errno( errno, "fork" );

    _pid = _working_pid;

    Z_LOG( "pid=" << _pid << "\n" );
}

//----------------------------------------------------------------------------Process::set_priority

bool Process::set_priority( int priority ) //, Has_log* log )
{
    try
    {
        if( _pid )
        {
            int error = setpriority( PRIO_PROCESS, _pid, priority );     // oder PRIO_PROCESS? S.a. getpriority()!
            if( error )  throw_errno( errno, "setpriority" );
        }
        else
        {
            _priority = priority;
            _priority_set = true;
        }
        
        return true;
    }
    catch( exception& x )
    {
        //if( log )  log->warn( x.what() );
        Z_LOG( x << '\n' );

        return false;
    }
}

//--------------------------------------------------------------------------------Process::priority

int Process::priority()
{
    errno = 0;

    int result = getpriority( PRIO_PROCESS, _pid );
    if( result == -1  &&  errno != 0 )  throw_errno( errno, "getpriority" );

    return result;
}

//----------------------------------------------------------------------Process::set_priority_class

bool Process::set_priority_class( const string& priority ) //, Has_log* log )
{
    return set_priority( priority_from_string( priority ) );
}

//--------------------------------------------------------------------------Process::priority_class
// Siehe auch priority_from_string()

string Process::priority_class()
{
    int priority = this->priority();

    if( priority >= idle_priority )  return "idle";
    if( priority <= high_priority )  return "high";
    
    if( priority >  0 )  return "below_normal";
    if( priority <  0 )  return "above_normal";
    
    return "normal";
}

//--------------------------------------------------------------------------Process::raise_priority
/*
bool Process::raise_priority( int difference )
{
    return set_priority( priority() - difference );
}
*/
//----------------------------------------------------------------------------Process::call_waitpid

bool Process::call_waitpid( bool wait )
{
    bool ok     = false;
    int  status = 0;

    if( wait )  Z_LOG( "pid=" << _working_pid << " waitpid() ...\n" );
    int ret = waitpid( _working_pid, &status, wait? 0 : WNOHANG );

    int err = ret == -1? errno : 0;
    if( wait )  Z_LOG( "pid=" << _working_pid << " waitpid() => " << ret << " status=" << printf_string("0x%x",ret) << "\n" );

    if( ret == _working_pid )
    {
        if( WIFEXITED  ( status ) )  _exit_code          = WEXITSTATUS( status ),  ok = true;
        if( WIFSIGNALED( status ) )  _termination_signal = WTERMSIG   ( status ),  ok = true;  

        _working_pid = 0;
        _terminated = true;
    }
    else
    {
        if( ret == -1 )
        {
            Z_LOG( "pid=" << _working_pid << " waitpid()  ERRNO-" << err << "  " << strerror(err) << "\n" );
            if( err == ECHILD )  _working_pid = 0;
        }
    }

    return ok;
}

//------------------------------------------------------------------------------------Process::wait

void Process::wait()
{
    if( _working_pid )
    {
        call_waitpid( true );
    }
}

//------------------------------------------------------------------------------------Process::wait

bool Process::wait( double seconds )
{
    if( _working_pid )
    {
        double until = double_from_gmtime() + seconds;
        struct timespec w;  w.tv_sec = 0, w.tv_nsec = 100*1000*1000;  // 0.1s

        for( int i = 0;; i++ )
        {
            //Z_LOG2( "joacim", i << ". call_waitpid(false)\n" );
            if( call_waitpid( false ) )  return true;
            if( seconds <= 0 )  break;
            
            if( i == 20 )  w.tv_sec = 1, w.tv_nsec = 0;  // 1s
            if( double_from_gmtime() >= until )  break;

            nanosleep( &w, NULL );
        }

        return false;
    }
    else
        return true;
}

//------------------------------------------------------------------------------Process::terminated

bool Process::terminated()
{
    if( _working_pid )  call_waitpid( false );
    return _working_pid == 0;
}

//-------------------------------------------------------------------------------Process::exit_code

int Process::exit_code()
{
    assert_terminated();

    return _exit_code;
}

//----------------------------------------------------------------------Process::termination_signal

int Process::termination_signal()
{
    assert_terminated();

    return _termination_signal;
}

//------------------------------------------------------------------------------------Process::kill

void Process::kill( int kill_signal )
{
    if( kill_signal == 0 )  kill_signal = SIGKILL;

    if( _working_pid )
    {
        Z_LOG( "pid=" << _working_pid << " kill(" << kill_signal << ")\n" );
        int err = ::kill( _working_pid, kill_signal );
        if( err )  throw_errno( errno, "kill" );
    }
}

//--------------------------------------------------------------------------------Process::try_kill

bool Process::try_kill( int kill_signal )
{
    if( kill_signal == 0 )  kill_signal = SIGKILL;

    bool result = false;


    if( _working_pid )
    {
        int err = ::kill( _working_pid, kill_signal );

        int errn = err? errno : 0;

        Log_ptr log;
        log << "pid=" << _working_pid << " kill(" << kill_signal << ")";
        if( err )  log << "  errno=" << errn << " " << strerror(errn);
             else  result = true;
        log << "\n";
    }

    return false;
}

//--------------------------------------------------------------------------------Process::obj_name
/*
string Process::obj_name()
{
    return "pid=" + as_string( _working_pid );
}
*/
//-------------------------------------------------------------------------------------------------

} //namespace posix
} //namespace zschimmer


#endif
